@@ -528,7 +528,7 @@ class TransformerDecoderLayer(nn.Module):
         tgt = tgt + self.dropout2(tgt2)
         tgt = self.norm2(tgt)
         tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
-        tgt = tgt + self.dropout3(tgt2)
+        tgt = (tgt + self.dropout3(tgt2))
         tgt = self.norm3(tgt)
         return tgt
 
